import time
import os
import tempfile
import logging


def ignore_exception(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return None
    return wrapper


def list_ray_logs(user_ray_dir: str, logger: logging.Logger = None):
    logger = logger if logger else logging.getLogger(__name__)
    logs = []
    if os.path.exists(user_ray_dir):
        dirs = ignore_exception(os.listdir)(user_ray_dir)
        if dirs is None:
            return logs
        for f in dirs:
            if f.startswith('session_'):
                log_dir = os.path.join(user_ray_dir, f, 'logs')
                if os.path.exists(log_dir):
                    log_dirs = ignore_exception(os.listdir)(log_dir)
                    if log_dirs is None:
                        continue
                    for log_file in log_dirs:
                        if "worker-" in log_file:
                            log_file_path = os.path.join(log_dir, log_file)
                            is_file = ignore_exception(os.path.isfile)(log_file_path)
                            if is_file is None:
                                continue
                            if is_file:
                                logs.append(log_file_path)
    return logs

def clean_ray_logs(logger: logging.Logger = None):
    logger = logger if logger else logging.getLogger(__name__)
    user_ray_dir = os.path.join(tempfile.gettempdir(), 'ray')
    size_cleaned = 0
    if os.path.exists(user_ray_dir):
        logger.info(f"Cleaning up Ray logs in {user_ray_dir}")
        all_logs = list_ray_logs(user_ray_dir)
        for log_file_path in all_logs:
            is_file = ignore_exception(os.path.isfile)(log_file_path)
            if is_file is None:
                continue
            if is_file:
                # Delete the log file if it is older than 16 minutes
                file_time = ignore_exception(os.path.getmtime)(log_file_path)
                if file_time is None:
                    continue
                if time.time() - file_time > 16 * 60:
                    file_size = ignore_exception(os.path.getsize)(log_file_path)
                    if file_size is None:
                        continue
                    ignore_exception(os.remove)(log_file_path)
                    size_cleaned += file_size
        logger.info(f"Cleaned up {size_cleaned/1024/1024} MB of Ray logs")

if __name__ == "__main__":
    # log to console
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    clean_ray_logs()


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import logging
import os
os.environ["HYDRA_FULL_ERROR"] = "1"
import ray
import hydra
import time
from ray.runtime_env import RuntimeEnv
from proof_wala.main.config import ExperimentType, parse_config
from proof_wala.main.run_training import train_experiment
from proof_wala.main.run_token_count import count_tokens

@hydra.main(config_path="config", config_name="experiment", version_base="1.2")
def main(cfg):
    print("To run this experiment on multiple GPUs, use the following command:")
    print(f"torchrun --nproc-per-node <num-proc-per-node> {__file__}")
    print(f"Working directory: {os.getcwd()}")
    print(f"Config: {cfg}")
    experiment = parse_config(cfg)
    time_now = time.strftime("%Y%m%d-%H%M%S")
    if experiment.expertiment_type == ExperimentType.Training:
        train_experiment(experiment, time_now)
    elif experiment.expertiment_type == ExperimentType.TokenCount:
        os.environ["COMET_MODE"] = "DISABLED"
        count_tokens(experiment)
    else:
        raise Exception(f"Invalid experiment type: {experiment.expertiment_type}")
    pass

def run():
    # Start the ray cluster
    os.environ["PYTHONPATH"] = f"{root_dir}:{os.environ.get('PYTHONPATH', '')}"
    environ = os.environ.copy()
    runtime_env = RuntimeEnv(
        env_vars=environ
    )
    ray.init(
        num_cpus=10, 
        object_store_memory=50*2**30, 
        _memory=50*2**30, 
        logging_level=logging.ERROR, 
        ignore_reinit_error=False, 
        log_to_driver=False, 
        configure_logging=False,
        _system_config={"metrics_report_interval_ms": 10**8},
        runtime_env=runtime_env)
    main()

if __name__ == "__main__":
    run()


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import logging
import copy
import os
import time
import enum
import uuid
from abc import ABC, abstractmethod
from itp_interface.tools.training_data_format import TrainingDataFormat, TrainingDataMetadataFormat
from itp_interface.tools.training_data import TrainingData
from itp_interface.rl.simple_proof_env import ProofEnv, ProofState, ProofAction, ProofTree, ProgressState
from itp_interface.rl.simpl_proof_env_pool import ProofEnvPool, replicate_proof_env
from itp_interface.tools.dynamic_coq_proof_exec import DynamicProofExecutor as DynamicCoqExecutor
from itp_interface.tools.dynamic_lean4_proof_exec import DynamicProofExecutor as DynamicLean4Executor
from itp_interface.rl.proof_tree import ProofSearchResult
from proof_wala.search.search import SearchAlgorithm, Node, Edge
from collections import namedtuple
from dataclasses import dataclass
from dataclasses_json import dataclass_json
 
# Declaring namedtuple()
ProofStateInfo = namedtuple('ProofStateInfo', ['proof_state', 'done', 'info', 'env_idx'])

class SearchPolicy(enum.Enum):
    DISCARD_FAILED_NODES = 1
    KEEP_FAILED_NODES = 2


@dataclass_json
@dataclass
class ProofPathTracer:
    collect_traces: bool
    folder: str
    training_meta_filename: str
    training_metadata: TrainingDataMetadataFormat
    max_parallelism: int

    def reset_training_data(self, logger: logging.Logger = None):
        time_str = time.strftime("%Y%m%d-%H%M%S")
        trace_folder = os.path.join(self.folder, time_str)
        os.makedirs(trace_folder, exist_ok=True)
        self.training_data = TrainingData(
            trace_folder,
            self.training_meta_filename, 
            self.training_metadata, 
            self.max_parallelism, 
            logger)
    
    def trace(self, training_data_format: TrainingDataFormat):
        if self.collect_traces and hasattr(self, 'training_data'):
            self.training_data.merge(training_data_format)
    
    def save(self):
        if self.collect_traces and hasattr(self, 'training_data'):
            self.training_data.save()

def convert_state_to_string(state: typing.Union[ProofState, TrainingDataFormat]) -> str:
    tdf = state.training_data_format if isinstance(state, ProofState) else state
    goals = tdf.start_goals
    if len(goals) == 0:
        return tdf.goal_description
    else:
        all_lines = []
        goal_len = len(goals)
        for idx, goal in enumerate(tdf.start_goals):
            all_lines.append(f"({idx+1}/{goal_len})")
            for hyp in goal.hypotheses:
                all_lines.append(hyp)
            if goal.goal is not None:
                all_lines.append(f"⊢ {goal.goal}")
        return '\n'.join(all_lines)

def end_state_string(language: ProofAction.Language) -> str:
    if language == ProofAction.Language.COQ:
        return DynamicCoqExecutor.NotInProofModeDescription
    elif language == ProofAction.Language.LEAN4:
        return DynamicLean4Executor.NotInProofModeDescription
    else:
        raise ValueError(f"Language {language} not supported")


class ProofActionGenerator(ABC):
    def __init__(self, width: int = 4):
        assert width > 0, "Width must be greater than 0"
        self.width = width
        pass
    
    @abstractmethod
    def get_proof_end_for_language(self, language: ProofAction.Language) -> ProofAction:
        pass

    @abstractmethod
    def generate_actions(self, state_info: ProofStateInfo, k: int = None) -> typing.List[typing.Tuple[float, ProofAction]]:
        pass

class ProofSearchBranchGenerator(ABC):
    def __init__(self, 
        width: int,
        envs : ProofEnvPool,
        env_to_state_map: typing.List[typing.Optional[int]],
        state_to_id_map: typing.Dict[str, int], 
        proof_action_generator: ProofActionGenerator,
        theorem_name: str,
        file_path: str,
        project_id: str,
        logger: logging.Logger,
        tracer: typing.Optional[ProofPathTracer] = None,
        original_proofs: typing.Optional[typing.List[typing.List[TrainingDataFormat]]] = None,
        search_policy: SearchPolicy = SearchPolicy.DISCARD_FAILED_NODES):
        assert proof_action_generator is not None, "Proof action generator cannot be None"
        assert envs is not None, "Environments cannot be None"
        assert envs.pool_size > 0, "Environments must contain at least one environment"
        self.width = width
        self.proof_action_generator = proof_action_generator
        self.envs = envs
        self.env_to_state_map = env_to_state_map
        self.state_id_to_env_map : typing.Dict[int, typing.Set[int]] = {}
        self.state_to_state_id_map = state_to_id_map
        self.tracer = tracer if tracer is not None else ProofPathTracer(False, "", "", TrainingDataMetadataFormat(), 1)
        self.original_proofs = original_proofs if original_proofs is not None else [[]]
        self.theorem_name = theorem_name
        self.file_path = file_path
        self.project_id = project_id
        self.original_proofs_state_names = set()
        self.state_action_map = set()
        self.logger = logger
        for proof in self.original_proofs:
            for state in proof:
                self.original_proofs_state_names.add(convert_state_to_string(state))
        for env_idx, state_idx in enumerate(env_to_state_map):
            if state_idx is not None and state_idx not in self.state_id_to_env_map:
                self.state_id_to_env_map[state_idx] = set()
            if state_idx is not None:
                self.state_id_to_env_map[state_idx].add(env_idx)
        self.search_policy = search_policy
        self.proof_id = os.path.join(str(uuid.uuid4()), self.project_id, self.theorem_name)
        pass

    def get_unused_envs(self) -> typing.List[int]:
        # Get all the environments for the start state
        return list(self.state_id_to_env_map[0]) if 0 in self.state_id_to_env_map else []
    
    def add_new_envs_to_pool(self, count: int):
        if count <= 0:
            return
        old_pool_size = self.envs.pool_size
        self.envs.add_and_init_proof_envs(count)
        for env_idx in range(old_pool_size, self.envs.pool_size):
            self.env_to_state_map.append(0)
            if 0 not in self.state_id_to_env_map:
                self.state_id_to_env_map[0] = set()
            self.state_id_to_env_map[0].add(env_idx)
        pass

    def reset_envs(self, env_idxs: typing.List[int], actions_till_states : typing.List[typing.List[ProofAction]], force_reset: bool = True) -> typing.Set[int]:
        if force_reset:
            if len(env_idxs) > 0:
                self.envs.reset(env_idxs)
        for env_idx in env_idxs:
            env_state_idx = self.env_to_state_map[env_idx]
            assert env_state_idx in self.state_id_to_env_map, f"Env state idx {env_state_idx} not found in the state id to env map"
            # assert env_idx in self.state_id_to_env_map[env_state_idx], f"Env idx {env_idx} not found in the state id to env map for state {env_state_idx}"
            if env_idx in self.state_id_to_env_map[env_state_idx]:
                self.state_id_to_env_map[env_state_idx].remove(env_idx)
        new_env_state_idx = None
        # Remove the env_idx from the old state id
        actions_zipped : typing.List[typing.List[ProofAction]]= []
        env_idxs_zipped : typing.List[typing.List[int]] = []
        for _idx, actions in enumerate(actions_till_states):
            assigned_env = env_idxs[_idx]
            for __idx, action in enumerate(actions):
                if len(actions_zipped) <= __idx:
                    actions_zipped.append([])
                if len(env_idxs_zipped) <= __idx:
                    env_idxs_zipped.append([])
                env_idxs_zipped[__idx].append(assigned_env)
                actions_zipped[__idx].append(action)
        erred_envs = set()
        erred_envs.update(self.envs.get_errd_envs())
        for actions_flat, env_idxs_flat in zip(actions_zipped, env_idxs_zipped):
            env_idxs_flat_with_idx = [(_idx, env_idx) for _idx, env_idx in enumerate(env_idxs_flat) if env_idx not in erred_envs]
            actions_flat = [actions_flat[_idx] for _idx, _ in env_idxs_flat_with_idx]
            env_idxs_flat = [env_idx for _, env_idx in env_idxs_flat_with_idx]
            if len(env_idxs_flat) > 0:
                self.envs.step(actions_flat, env_idxs_flat)
            new_erred_envs = self.envs.get_errd_envs()
            erred_envs.update(new_erred_envs)
        if len(erred_envs) > 0:
            # Remove the envs from the state id to env map
            for env_idx in erred_envs:
                env_state_idx = self.env_to_state_map[env_idx]
                if env_state_idx is not None and env_idx in self.state_id_to_env_map[env_state_idx]:
                    self.state_id_to_env_map[env_state_idx].remove(env_idx)
                self.env_to_state_map[env_idx] = None
        safe_env_idxs = [env_idx for env_idx in env_idxs if env_idx not in erred_envs]
        if len(safe_env_idxs) > 0:
            safe_env_states = self.envs.get_state(safe_env_idxs)
        else:
            safe_env_states = []
        env_states = [None for _ in range(len(env_idxs))]
        for idx, env_idx in enumerate(safe_env_idxs):
            actual_env_idx = env_idxs.index(env_idx)
            env_states[actual_env_idx] = safe_env_states[idx]
        for env_state, env_idx in zip(env_states, env_idxs):
            if env_state is None:
                continue # This means that the environment failed and we should ignore it
            env_state_str = convert_state_to_string(env_state)
            if env_state_str not in self.state_to_state_id_map:
                new_env_state_idx = len(self.state_to_state_id_map)
                self.state_to_state_id_map[env_state_str] = new_env_state_idx
            else:
                new_env_state_idx = self.state_to_state_id_map[env_state_str]
            self.env_to_state_map[env_idx] = new_env_state_idx
            if new_env_state_idx not in self.state_id_to_env_map:
                self.state_id_to_env_map[new_env_state_idx] = set()
            self.state_id_to_env_map[new_env_state_idx].add(env_idx)
        return erred_envs

    def reclaim_envs(self, env_idxs: typing.List[int], language: ProofAction.Language):
        if len(env_idxs) > 0 and language == ProofAction.Language.COQ:
            action_per_env = []
            for env_idx in env_idxs:
                env_state_idx = self.env_to_state_map[env_idx]
                backtrack_actions = [ProofAction(ProofAction.ActionType.BACKTRACK, language) for _ in range(env_state_idx)]
                action_per_env.append(backtrack_actions)
            self.logger.info(f"Reclaiming {len(env_idxs)} environments")
            _t_start = time.time()
            self.reset_envs(env_idxs, action_per_env, force_reset=False)
            _t_end = time.time()
            self.logger.info(f"Reclaimed {len(env_idxs)} environments in {_t_end - _t_start} seconds")

    def _check_if_within_timeout(self, start_time: float, timeout_in_secs: float, logger: logging.Logger) -> bool:
        timedout = time.time() - start_time >= timeout_in_secs
        if timedout:
            logger.info("Search timed out, returning without generating more nodes")
        return timedout

    def _update_state_to_env_map(self, next_state, done, info, env_idxs, idx, actions, node, actions_to_run, state, state_idx):
        state_name = convert_state_to_string(next_state) if next_state is not None else 'Tactic failed/timed out'
        action_str = '\n'.join(actions[idx].kwargs['tactics']) if actions[idx].kwargs['tactics'] is not None else str(actions[idx])
        self.logger.info(f"Action: {action_str}, Done: {done}, Progress: {info.progress if info is not None else None}, Error: {info.error_message if info is not None else ('Tactic failed/timed out' if next_state is None else None)}")
        new_state_id = None
        if state_name not in self.state_to_state_id_map:
            new_state_id = len(self.state_to_state_id_map)
            self.state_to_state_id_map[state_name] = new_state_id
        else:
            new_state_id = self.state_to_state_id_map[state_name]
        self.env_to_state_map[env_idxs[idx]] = new_state_id
        # Remove the env_idx from the old state id
        if env_idxs[idx] in self.state_id_to_env_map[state_idx]:
            self.state_id_to_env_map[state_idx].remove(env_idxs[idx])
        if new_state_id not in self.state_id_to_env_map:
            self.state_id_to_env_map[new_state_id] = set()
        self.state_id_to_env_map[new_state_id].add(env_idxs[idx])
        if (state_name, action_str) not in self.state_action_map and self.tracer.collect_traces:
            self.state_action_map.add((state_name, action_str))
            on_proof_path = state_name in self.original_proofs_state_names
            additional_info = {
                'done': done,
                'progress': info.progress if info is not None else None,
                'error_message': info.error_message if info is not None else ('Tactic failed/timed out' if next_state is None else None),
                'distance_from_root': node.distance_from_root + 1,
                'score': actions_to_run[idx][1],
                'on_proof_path': on_proof_path,
                'start_goal_id': state_idx,
                'end_goal_id': new_state_id
            }
            training_data_format = TrainingDataFormat(
                proof_id=self.proof_id,
                goal_description=next_state.training_data_format.goal_description if next_state is not None else 'Tactic failed/timed out',
                start_goals=state.training_data_format.start_goals,
                proof_steps=copy.deepcopy(actions[idx].kwargs['tactics']),
                end_goals=next_state.training_data_format.start_goals if next_state is not None else state.training_data_format.start_goals,
                addition_state_info=additional_info,
                file_path=self.file_path,
                project_id=self.project_id,
                theorem_name=self.theorem_name
            )
            self.tracer.trace(training_data_format)
        return new_state_id, state_name

    def __call__(self, node: Node, timeout_in_secs: float) -> typing.Tuple[typing.List[Node], typing.List[Edge]]:
        # Call the proof action generator to generate actions for each environment
        if node.other_data is None or timeout_in_secs <= 0:
            return [], [] # This is kind of dead end
        # for _key, _valu in self.state_to_state_id_map.items():
        #     self.logger.info(f"State_to_state_id_map: {_valu} -> \n{_key}")
        # for _key, _valu in self.state_id_to_env_map.items():
        #     self.logger.info(f"State_id_to_env_map: {_key} -> {_valu}")
        node_gen_start_time = time.time()
        state_info : ProofStateInfo = node.other_data
        state: ProofState = state_info.proof_state
        if state_info.done:
            return [], []
        state_str = convert_state_to_string(state)
        state_idx = self.state_to_state_id_map.get(state_str, -1)
        assert state_idx != -1, f"State {state_str} not found in the state to state id map"
        actions_scores = self.proof_action_generator.generate_actions(state_info, k=self.width)
        nodes = []
        edges = []
        if self._check_if_within_timeout(node_gen_start_time, timeout_in_secs, self.logger):
            return nodes, edges
        while len(actions_scores) > 0:
            env_idxs : typing.List[int] = list(self.state_id_to_env_map.get(state_idx, set()))
            if len(env_idxs) < len(actions_scores):
                # This means that there is some backtracking and somewhere the original path got lost
                # Reset the environment to the state
                env_idx: int = state_info.env_idx
                proof_tree : ProofTree = state.proof_tree
                actions_till_state : typing.List[ProofAction] = proof_tree.actions
                free_envs = self.get_unused_envs()
                diff = len(actions_scores) - len(env_idxs)
                if len(free_envs) < diff:
                    self.logger.info(f"Adding {diff - len(free_envs)} new environments to the pool")
                    _t_start = time.time()
                    self.add_new_envs_to_pool(diff - len(free_envs))
                    _t_end = time.time()
                    self.logger.info(f"Added {diff - len(free_envs)} new environments in {_t_end - _t_start} seconds")
                    free_envs = self.get_unused_envs()
                free_envs = free_envs[:diff]
                self.logger.info(f"Resetting {len(free_envs)} environments")
                _t_start = time.time()
                erred_envs = self.reset_envs(free_envs, [actions_till_state for _ in range(len(free_envs))], force_reset=False)
                # Remove the erred envs from the free envs
                free_envs = [env_idx for env_idx in free_envs if env_idx not in erred_envs]
                _t_end = time.time()
                self.logger.info(f"Reset {len(free_envs)} environments in {_t_end - _t_start} seconds")
                env_idxs.extend(free_envs)
            assert len(env_idxs) > 0, f"No environments found for state {state_str}"

            actions_to_run = []
            for _idx in range(min(len(env_idxs), len(actions_scores))):
                env_idx = env_idxs[_idx]
                score, action = actions_scores.pop()
                actions_to_run.append((env_idx, score, action))
            env_idxs = [env_idx for env_idx, _, _ in actions_to_run]
            actions = [action for _, _, action in actions_to_run]
            action_start_time = time.time()
            if len(env_idxs) > 0:
                results = self.envs.step(actions, env_idxs)
            else:
                results = []
            action_end_time = time.time()
            self.logger.info(f"Finished executing {len(actions)} actions parallely in {action_end_time - action_start_time} seconds.")
            failed_envs = []
            for idx, result in enumerate(results):
                _, _, next_state, _, done, info  = result
                proof_state_info = ProofStateInfo(next_state, done, info, env_idxs[idx])
                new_state_id, state_name = self._update_state_to_env_map(next_state, done, info, env_idxs, idx, actions, node, actions_to_run, state, state_idx)
                
                # TODO: Even though Lean 4 does not have an explicit Qed tactic, 
                # we should still use `done` to check if the proof is complete.
                # This will require changes in the itp_interface Lean4 code, 
                # to not finish the proof until the Qed tactic explicitly stated i.e. 
                # `done` is explicitly used to signal the end of the proof.
                # This will be automatically achieved as soon as https://github.com/trishullab/itp-interface/issues/31 is resolved.
                if next_state is not None and len(next_state.training_data_format.start_goals) == 0 and not done:
                    # We found a very good action so we should signal the search to stop, regardless of the search heuristic
                    _temp_list = list(actions_to_run[idx])
                    _temp_list[1] = -100
                    actions_to_run[idx] = tuple(_temp_list)
                    # Found a proof, something like Qed
                    qed_tactic = self.proof_action_generator.get_proof_end_for_language(next_state.language)
                    result = self.envs.step([qed_tactic], [env_idxs[idx]])[0]
                    _, _, next_next_state, _, done, info  = result
                    actions[idx].kwargs['tactics'].extend(qed_tactic.kwargs['tactics'])
                    proof_state_info = ProofStateInfo(next_next_state, done, info, env_idxs[idx])
                    new_state_id, state_name = self._update_state_to_env_map(next_next_state, done, info, [env_idxs[idx]], 0, [qed_tactic], node, [(env_idxs[idx], -100, qed_tactic)], next_state, new_state_id)
                    next_state = next_next_state

                # TODO: Don't add the node if the action failed and the search policy is to discard failed nodes
                if (info is not None and info.progress != ProgressState.FAILED and info.progress != ProgressState.STATE_UNCHANGED) or \
                    self.search_policy != SearchPolicy.DISCARD_FAILED_NODES:
                    edges.append(Edge('\n'.join(actions[idx].kwargs['tactics']), actions_to_run[idx][1], actions_to_run[idx][2]))
                    nodes.append(Node(state_name, actions_to_run[idx][1], proof_state_info))
                else:
                    failed_envs.append(env_idxs[idx])   
            self.reclaim_envs(failed_envs, state.language)
            if self._check_if_within_timeout(node_gen_start_time, timeout_in_secs, self.logger):
                return nodes, edges
        node_gen_end_time = time.time()
        self.logger.info(f"Finished executing {len(nodes)} branches in {node_gen_end_time - node_gen_start_time} seconds")
        return nodes, edges
    
class ProofSearhHeuristic(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, node: Node) -> float:
        pass

class ProofSearchDriver:
    def __init__(self, 
        search_algorithm: SearchAlgorithm,
        proof_action_generator: ProofActionGenerator,
        proof_search_heuristic: ProofSearhHeuristic,
        width: int = 4,
        logger: logging.Logger = None,
        tracer: typing.Optional[ProofPathTracer] = None,
        search_policy: SearchPolicy = SearchPolicy.DISCARD_FAILED_NODES):
        assert search_algorithm is not None, "Search algorithm cannot be None"
        assert proof_action_generator is not None, "Proof action generator cannot be None"
        assert proof_search_heuristic is not None, "Proof search heuristic cannot be None"
        assert width > 0, "Width must be greater than 0"
        self.search_algorithm = search_algorithm
        self.proof_action_generator = proof_action_generator
        self.logger = logger if logger is not None else logging.getLogger(__name__)
        self.width = width
        self.proof_search_heuristic = proof_search_heuristic
        # don go beyond 0.6 * os.cpu_count()
        max_parallelism = int(0.15625 * os.cpu_count())
        self.env_count = max(min(8 * self.width, max_parallelism), 1)   # We need more environments to run in parallel without waiting
        self.tracer = tracer if tracer is not None else ProofPathTracer(False, "", "", TrainingDataMetadataFormat(), 1)
        self.search_policy = search_policy

    def search_proof(
        self, 
        env: ProofEnv, 
        timeout_in_secs: int = 60,
        dump_file_name: str = None,
        additional_info: dict = {},
        extract_original_proofs: bool = False,
        original_proofs: typing.Optional[typing.List[typing.List[TrainingDataFormat]]] = None
    ) -> typing.Tuple[Node, Node, ProofSearchResult]:
        if extract_original_proofs and original_proofs is None:
            env_cpy = replicate_proof_env(env)
            env_cpy1 = replicate_proof_env(env)
            original_proof = []
            with env_cpy:
                with env_cpy1:
                    done = False
                    while not done:
                        env_cpy._dynamic_proof_executor.run_next()
                        stmt = env_cpy._dynamic_proof_executor.current_stmt
                        proof_steps = [stmt]
                        tdf = env_cpy1.state.training_data_format
                        res = env_cpy1.step(
                            ProofAction(
                            ProofAction.ActionType.RUN_TACTIC, 
                            env_cpy.language, 
                            tactics=proof_steps))
                        _, _, _, _, done, _  = res
                        if not done:
                            tdf.proof_steps = proof_steps
                            original_proof.append(tdf)
            original_proofs = [original_proof]

        pool = ProofEnvPool(self.env_count, proof_env=env, logger=self.logger, timeout=timeout_in_secs, max_parallel_envs=self.env_count)
        start_goal = None
        tree_node = None
        proof_search_res = None
        full_path = env.dynamic_proof_executor_callback.file_path
        language = env.language
        found = False
        got_exception = False
        theorem_name = env.lemma_name
        lemma_name = theorem_name
        try:
            pool.__enter__()
            erred_envs = pool.get_errd_envs()
            stable_envs = [env_idx for env_idx in range(self.env_count) if env_idx not in erred_envs]
            if len(stable_envs) == 0:
                self.logger.error("All environments failed to initialize, returning None")
                proof_search_res = ProofSearchResult(
                    full_path, 
                    False, 
                    env.lemma_name, 
                    [], 
                    10**6, 
                    -1, 
                    possible_failed_paths=-1, 
                    num_of_backtracks=-1, 
                    is_timeout=True, 
                    is_inference_exhausted=False, 
                    longest_success_path=-1,
                    additional_info=additional_info,
                    language=language)
            else:
                start_state = pool.get_state([stable_envs[0]])[0]
                start_state_info = ProofStateInfo(start_state, False, None, 0)
                env_to_state_map = [0 if idx in stable_envs else None for idx in range(self.env_count)]
                start_state_str = convert_state_to_string(start_state)
                state_to_id_map = {start_state_str: stable_envs[0]}
                # the lower the score the better
                start_goal = Node(start_state_str, 100, start_state_info)
                language : ProofAction.Language = pool._get_attr('language', [stable_envs[0]])[0]
                end_goal = Node(end_state_string(language), -100)
                file_path = env.dynamic_proof_executor_callback.file_path
                project_id = env.dynamic_proof_executor_callback.project_folder 
                branch_generator = ProofSearchBranchGenerator(
                    self.width, 
                    pool, 
                    env_to_state_map, 
                    state_to_id_map, 
                    self.proof_action_generator, 
                    theorem_name,
                    file_path,
                    project_id,
                    self.logger,
                    self.tracer,
                    original_proofs)
                temp_start_time = time.time()
                try:
                    tree_node, found, time_taken = self.search_algorithm.search(
                        start_goal,
                        end_goal,
                        self.proof_search_heuristic, 
                        branch_generator, 
                        parallel_count=self.width,
                        timeout_in_secs=timeout_in_secs,
                        logger=self.logger)
                except Exception as e:
                    got_exception = True
                    tree_node = None
                    found = False
                    time_taken = time.time() - temp_start_time
                    self.logger.error(f"Search failed with error: {e}")
                    proof_search_res = ProofSearchResult(
                        full_path, 
                        False, 
                        lemma_name, 
                        [], 
                        time_taken, 
                        -1, 
                        possible_failed_paths=-1, 
                        num_of_backtracks=-1, 
                        is_timeout=time_taken >= timeout_in_secs, 
                        is_inference_exhausted=False, 
                        longest_success_path=-1,
                        additional_info=additional_info,
                        language=language)
                    self.logger.warning(f"Going to dump the proof search result")
                    if dump_file_name is not None:
                        opening_mode = 'a' if os.path.exists(dump_file_name) else 'w'
                        with open(dump_file_name, opening_mode) as f:
                            if opening_mode == 'a':
                                f.write("\n\n")
                            f.write(str(proof_search_res))
                    self.logger.warning(f"Returning the proof search result")
                    return start_goal, tree_node, proof_search_res

                if found and not got_exception:
                    erred_envs = pool.get_errd_envs()
                    stable_envs = [env_idx for env_idx in range(self.env_count) if env_idx not in erred_envs]
                    try:
                        lemma_name = pool._get_attr('_lemma_name_with_stmt', [stable_envs[0]])[0]
                    except Exception as e:
                        lemma_name = theorem_name
                else:
                    lemma_name = theorem_name
                if not got_exception:
                    pool.__exit__(None, None, None)
                # full_path = env.dynamic_proof_executor_callback.file_path
        except Exception as e:
            self.logger.error(f"Proof search failed with error: {e}")
        if found:
            last_state: ProofState = tree_node.other_data.proof_state
            proof_tree: ProofTree = last_state.proof_tree
            actions_till_state: typing.List[ProofAction] = proof_tree.tactics
            proof_steps = [TrainingDataFormat(proof_steps=tactic.proof_steps) for _, tactic in actions_till_state]
            proof_search_res = ProofSearchResult(
                        full_path,
                        True, 
                        lemma_name,
                        proof_steps, 
                        time_taken, 
                        -1, 
                        possible_failed_paths=-1, 
                        num_of_backtracks=-1, 
                        is_timeout=False, 
                        is_inference_exhausted=False, 
                        longest_success_path=-1,
                        additional_info=additional_info,
                        language=language)
        elif proof_search_res is None:
            proof_search_res = ProofSearchResult(
                full_path, 
                False, 
                lemma_name, 
                [], 
                time_taken, 
                -1, 
                possible_failed_paths=-1, 
                num_of_backtracks=-1, 
                is_timeout=time_taken >= timeout_in_secs, 
                is_inference_exhausted=False, 
                longest_success_path=-1,
                additional_info=additional_info,
                language=language)

        if got_exception:
            self.logger.error("Got exception while searching for proof")
        else:
            self.logger.info("Did not get exception while searching for proof")

        if dump_file_name is not None:
            opening_mode = 'a' if os.path.exists(dump_file_name) else 'w'
            with open(dump_file_name, opening_mode) as f:
                if opening_mode == 'a':
                    f.write("\n\n")
                f.write(str(proof_search_res))
        return start_goal, tree_node, proof_search_res


import typing
import copy
from enum import Enum
from itp_interface.tools.training_data_format import TrainingDataFormat, Goal
try:
    from .grammars.grammar import Grammar
except ImportError:
    from grammars.grammar import Grammar


class CustomGoalHash(object):
    def __init__(self, goal: Goal):
        self.goal = goal

    def __hash__(self):
        return hash(self.goal.goal)

    def __eq__(self, other):
        if not isinstance(other, CustomGoalHash):
            return False
        return self.goal == other.goal

def get_delta_state(start_goals: typing.List[Goal], end_goals: typing.List[Goal]) -> typing.List[Goal]:
    delta_goals = []
    start_goals : typing.List[CustomGoalHash] = [CustomGoalHash(goal) for goal in start_goals]
    end_goals : typing.List[CustomGoalHash] = [CustomGoalHash(goal) for goal in end_goals]
    start_goal_set = set(start_goals)
    end_goal_set = set(end_goals)
    new_goals = end_goal_set.difference(start_goal_set)
    new_goal_ordered = [(idx, True, goal.goal) for idx, goal in enumerate(end_goals) if goal in new_goals]
    same_goals = start_goal_set.intersection(end_goal_set)
    same_goal_ordered = [(idx, False, goal.goal) for idx, goal in enumerate(end_goals) if goal in same_goals]
    all_goals_ordered = same_goal_ordered + new_goal_ordered
    # Sort the goals based on the order of the goals in the end_goals
    all_goals_ordered = sorted(all_goals_ordered, key=lambda x: x[0])
    delta_goals = []
    for _, is_new, goal in all_goals_ordered:
        if is_new:
            delta_goals.append(goal)
        else:
            goal_idx = start_goals.index(CustomGoalHash(goal))
            assert goal_idx >= 0 and goal_idx < len(start_goals), f"Goal index {goal_idx} is out of bounds"
            delta_goals.append(Goal(goal=f"DITTO {goal_idx}", hypotheses=[]))
    return delta_goals

def get_full_state(start_goals: typing.List[Goal], delta_goals: typing.List[Goal]) -> typing.List[Goal]:
    full_goals = []
    for goal in delta_goals:
        if goal.goal.startswith("DITTO"):
            idx = int(goal.goal.split(" ")[-1])
            full_goals.append(start_goals[idx])
        else:
            full_goals.append(goal)
    return full_goals


class ProofModelGrammar(Grammar):
    grammar = """
Prog:
   StatesStr ProofState End
|  StatesStr ProofState ProofStep End;
ProofState:
  Description String States
| States;
States:
  State
| State States
| EMPTY;
State:
 StateNum String HypsResponses;
HypsResponses:
    Hyps HypResponses
|   EMPTY;
HypResponses:
  HypResponse
| HypResponse HypResponses;
HypResponse:
  Hyp String;
ProofStep: 
ProofStepStr String;
StateNum:
    StateStr int;


terminals
int: /\d+/;
StateStr: "[STATE]";
StatesStr: "[STATES]";
ProofStepStr: "[PROOFSTEP]";
Hyps: "[HYPOTHESES]";
Hyp: "[HYPOTHESIS]";
End: "[END]";
Description: "[DESCRIPTION]";
String:;
"""
    class Keywords(Enum):
        STATE = "[STATE]"
        STATES = "[STATES]"
        PROOFSTEP = "[PROOFSTEP]"
        HYPOTHESES = "[HYPOTHESES]"
        HYPOTHESIS = "[HYPOTHESIS]"
        SUCCESS = "[SUCCESS]"
        END = "[END]"
        DESCRIPTION = "[DESCRIPTION]"

        def __str__(self) -> str:
            return self.value

    keywords = [keyword.value for keyword in Keywords]

    def before_keyword(text, pos):
        last = pos
        while last < len(text):
          while last < len(text) and text[last] != '[':
            last += 1
          if last < len(text):
              for keyword in ProofModelGrammar.keywords:
                  if text[last:].startswith(keyword):
                      return text[pos:last]
              last += 1

    def __init__(self, state_delta: bool = False, enable_proofstep: bool = True, ordering: typing.List[str] = ["DESCRIPTION", "STATE", "PROOFSTEP"]):
        recognizers = {
            'String': ProofModelGrammar.before_keyword
        }
        self.state_delta = state_delta
        self.enable_proofstep = enable_proofstep
        self.ordering = ordering
        super(ProofModelGrammar, self).__init__(ProofModelGrammar.grammar, ProofModelGrammar.keywords, recognizers=recognizers)
    
    def format_as_per_grammar(self, training_data_format: TrainingDataFormat, k: typing.Optional[int] = None, max_token_cnt: typing.Optional[int] = None, characters_per_token: float = 4.0) -> str:
        # Add algorithm for trimming the right amount of goals, theorems and defintions, steps, etc. based on the max_token_cnt
        char_cnt = int(max_token_cnt * characters_per_token) if max_token_cnt is not None else None # 4 is the average length of a token as per OpenAI
        text = ""
        lines_map = {
            ProofModelGrammar.Keywords.DESCRIPTION : [],
            ProofModelGrammar.Keywords.STATE : [],
            ProofModelGrammar.Keywords.PROOFSTEP : [],
        }
        lines_order = []
        try:
            for keyword in self.ordering:
                keyword_enum = ProofModelGrammar.Keywords[str(keyword)]
                if keyword_enum not in lines_order:
                    lines_order.append(keyword_enum)
        except:
            lines_order = [
                ProofModelGrammar.Keywords.DESCRIPTION,
                ProofModelGrammar.Keywords.STATE,
                ProofModelGrammar.Keywords.PROOFSTEP,
            ]
        priority_order_lo_hi = [
            ProofModelGrammar.Keywords.STATE,
            ProofModelGrammar.Keywords.PROOFSTEP,
        ]
        lines_map[ProofModelGrammar.Keywords.STATE] = [str(ProofModelGrammar.Keywords.STATES)]
        if training_data_format.goal_description is not None:
            new_line = f"{ProofModelGrammar.Keywords.DESCRIPTION}\n{training_data_format.goal_description}\n"
            lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
        for i, goal in enumerate(training_data_format.start_goals):
            new_line = f"{ProofModelGrammar.Keywords.STATE} {i}" # Add the goal number
            lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
            new_line = str(goal.goal)
            lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
            if len(goal.hypotheses) > 0:
                new_line = f"{ProofModelGrammar.Keywords.HYPOTHESES}"
                lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
                for hyp in goal.hypotheses:
                    new_line = f"{ProofModelGrammar.Keywords.HYPOTHESIS} {hyp}"
                    lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
        if len(training_data_format.proof_steps) > 0 and self.enable_proofstep:
            new_line = f"\n{ProofModelGrammar.Keywords.PROOFSTEP}"
            lines_map[ProofModelGrammar.Keywords.PROOFSTEP] = [new_line]
            lines_map[ProofModelGrammar.Keywords.PROOFSTEP].extend(training_data_format.proof_steps)
        keywords = [keyword for keyword in lines_map.keys()]
        # Convert all the lines under each keyword to a single string
        for keyword in keywords:
            lines_map[keyword] = "\n".join(lines_map[keyword])
        # Frame the first prompt version without any token limit
        text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{ProofModelGrammar.Keywords.END}"
        
        # Trim the lines based on the max_token_cnt
        if char_cnt is not None and len(text) > char_cnt:
            _idx = 0
            diff = len(text) - char_cnt
            while _idx < len(priority_order_lo_hi) and diff > 0:
                trim_part = priority_order_lo_hi[_idx]
                if trim_part in lines_map:
                    lines_map[trim_part] = lines_map[trim_part][:-diff] # Trim everything except the STEPS from the end
                text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{ProofModelGrammar.Keywords.END}"
                diff = len(text) - char_cnt
                _idx += 1

        # verify that the text is valid as per grammar by compiling it
        # self.compile(text)
        if char_cnt is not None:
            assert len(text) <= char_cnt, f"Text length {len(text)} is greater than the max token count {char_cnt}. Possibly too few characters per token." +\
            f" characters_per_token = {characters_per_token}, max_token_cnt = {max_token_cnt}"
            # text = text[:char_cnt] # Just trim the text from the end because no trimming strategy has worked out
        return text

    def _parse_expr(self, rule_type, nodes, context: TrainingDataFormat):
        if rule_type == ProofModelGrammar.Keywords.HYPOTHESIS:
            hyps = []
            for node in nodes:
                hyps.extend(node)
            return hyps
        elif rule_type == ProofModelGrammar.Keywords.HYPOTHESES:
            return nodes[1] if len(nodes) > 1 else []
        elif rule_type == ProofModelGrammar.Keywords.STATE:
            hyps = nodes[2] if len(nodes) > 2 else []
            goal = nodes[1].strip() if len(nodes) > 1 else None
            state = Goal(hypotheses=hyps, goal=goal)
            return [state]
        elif rule_type == ProofModelGrammar.Keywords.STATES:
            states = []
            for node in nodes:
                states.extend(node)
            return states
        elif rule_type == "Prog":
            if len(nodes) >= 1 and nodes[0] == ProofModelGrammar.Keywords.DESCRIPTION.value:
                context.goal_description = nodes[1].strip() if nodes[1] is not None else None
                context.start_goals = nodes[2]
            else:
                context.start_goals = nodes
            return context
        else:
            raise Exception("Not implemented")

    def get_action(self, tdf):
        context = tdf
        actions = {
            "Prog": lambda _, nodes: self._parse_expr("Prog", nodes[1], context),
            "States": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.STATES, nodes, context),
            "State": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.STATE, nodes, context),
            "HypsResponses": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.HYPOTHESES, nodes, context),
            "HypResponses": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.HYPOTHESIS, nodes, context),
            "HypResponse": lambda _, nodes: [nodes[1]],
            "String": lambda _, nodes: str(nodes).strip(),
        }
        return actions

    def interpret_result(self, result):
        return result
    
    def parse(self, text, tdf: TrainingDataFormat = None) -> TrainingDataFormat:
        tdf = TrainingDataFormat() if tdf is None else copy.deepcopy(tdf)
        final_tdf : TrainingDataFormat = self.run(text, tdf)
        return final_tdf


class ProofModelPredGrammar(Grammar):
    grammar = """
Prog:
   ProofStep StatesStr ProofState End
|  StatesStr ProofState End;
ProofState:
  Description String States
| States;
States:
  State
| State States
| EMPTY;
State:
 StateNum String HypsResponses;
HypsResponses:
    Hyps HypResponses
|   EMPTY;
HypResponses:
  HypResponse
| HypResponse HypResponses;
HypResponse:
  Hyp String;
StateNum:
    StateStr int;
ProofStep:
    ProofStepStr String;

terminals
int: /\d+/;
StateStr: "[STATE]";
StatesStr: "[STATES]";
Hyps: "[HYPOTHESES]";
Hyp: "[HYPOTHESIS]";
End: "[END]";
Description: "[DESCRIPTION]";
ProofStepStr: "[PROOFSTEP]";
String:;
"""

    def __init__(self, state_delta: bool = False, enable_proofstep: bool = False, ordering: typing.List[str] = ["PROOFSTEP", "DESCRIPTION", "STATE"]):
        recognizers = {
            'String': ProofModelGrammar.before_keyword
        }
        self.state_delta = state_delta
        self.enable_proofstep = enable_proofstep
        self.ordering = ordering
        super(ProofModelPredGrammar, self).__init__(ProofModelPredGrammar.grammar, ProofModelGrammar.keywords, recognizers=recognizers)
    
    def format_as_per_grammar(self, training_data_format: TrainingDataFormat, k: typing.Optional[int] = None, max_token_cnt: typing.Optional[int] = None, characters_per_token: float = 4.0) -> str:
        # Add algorithm for trimming the right amount of goals, theorems and defintions, steps, etc. based on the max_token_cnt
        char_cnt = int(max_token_cnt * characters_per_token) if max_token_cnt is not None else None # 4 is the average length of a token as per OpenAI
        text = ""
        if self.enable_proofstep:
            lines_map = {
                ProofModelGrammar.Keywords.PROOFSTEP : [],
                ProofModelGrammar.Keywords.DESCRIPTION : [],
                ProofModelGrammar.Keywords.STATE : []
            }
            lines_order = []
            try:
                for keyword in self.ordering:
                    keyword_enum = ProofModelGrammar.Keywords[str(keyword)]
                    if keyword_enum not in lines_order:
                        lines_order.append(keyword_enum)
            except:
                lines_order = [
                    ProofModelGrammar.Keywords.PROOFSTEP,
                    ProofModelGrammar.Keywords.DESCRIPTION,
                    ProofModelGrammar.Keywords.STATE
                ]
            priority_order_lo_hi = [
                ProofModelGrammar.Keywords.PROOFSTEP,
                ProofModelGrammar.Keywords.DESCRIPTION,
                ProofModelGrammar.Keywords.STATE,
            ]
        else:
            lines_map = {
                ProofModelGrammar.Keywords.DESCRIPTION : [],
                ProofModelGrammar.Keywords.STATE : []
            }
            lines_order = []
            try:
                for keyword in self.ordering:
                    keyword_enum = ProofModelGrammar.Keywords[str(keyword)]
                    if keyword_enum not in lines_map:
                        lines_order.append(keyword_enum)
            except:
                lines_order = [
                    ProofModelGrammar.Keywords.DESCRIPTION,
                    ProofModelGrammar.Keywords.STATE
                ]
            priority_order_lo_hi = [
                ProofModelGrammar.Keywords.DESCRIPTION,
                ProofModelGrammar.Keywords.STATE,
            ]
        lines_map[ProofModelGrammar.Keywords.STATE] = [str(ProofModelGrammar.Keywords.STATES)]
        if len(training_data_format.proof_steps) > 0 and self.enable_proofstep:
            new_line = f"{ProofModelGrammar.Keywords.PROOFSTEP}"
            lines_map[ProofModelGrammar.Keywords.PROOFSTEP].append(new_line)
            lines_map[ProofModelGrammar.Keywords.PROOFSTEP].extend(training_data_format.proof_steps)
        additional_info = ""
        if training_data_format.addition_state_info is not None:
            if len(training_data_format.addition_state_info) > 0:
                progress = training_data_format.addition_state_info.get("progress", None)
                if progress is not None:
                    additional_info += f"Progress: {progress}\n"
                error_message = training_data_format.addition_state_info.get("error_message", None)
                if error_message is not None:
                    additional_info += f"Error: {error_message}\n"
        if training_data_format.goal_description is not None or len(additional_info) > 0:
            if training_data_format.goal_description is not None:
                new_line = f"{ProofModelGrammar.Keywords.DESCRIPTION}\n{training_data_format.goal_description}\n"
            else:
                new_line = f"{ProofModelGrammar.Keywords.DESCRIPTION}\n"
            new_line += additional_info
            lines_map[ProofModelGrammar.Keywords.DESCRIPTION].append(new_line)
        end_goals = training_data_format.end_goals if not self.state_delta else get_delta_state(training_data_format.start_goals, training_data_format.end_goals)
        for i, goal in enumerate(end_goals):
            new_line = f"{ProofModelGrammar.Keywords.STATE} {i}"
            lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
            new_line = str(goal.goal)
            lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
            if len(goal.hypotheses) > 0:
                new_line = f"{ProofModelGrammar.Keywords.HYPOTHESES}"
                lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
                for hyp in goal.hypotheses:
                    new_line = f"{ProofModelGrammar.Keywords.HYPOTHESIS} {hyp}"
                    lines_map[ProofModelGrammar.Keywords.STATE].append(new_line)
        keywords = [keyword for keyword in lines_map.keys()]
        # Convert all the lines under each keyword to a single string
        for keyword in keywords:
            lines_map[keyword] = "\n".join(lines_map[keyword])
        # Frame the first prompt version without any token limit
        text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{ProofModelGrammar.Keywords.END}"
        
        # Trim the lines based on the max_token_cnt
        if char_cnt is not None and len(text) > char_cnt:
            _idx = 0
            diff = len(text) - char_cnt
            while _idx < len(priority_order_lo_hi) and diff > 0:
                trim_part = priority_order_lo_hi[_idx]
                if trim_part in lines_map:
                    lines_map[trim_part] = lines_map[trim_part][:-diff] # Trim everything except the STEPS from the end
                text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{ProofModelGrammar.Keywords.END}"
                diff = len(text) - char_cnt
                _idx += 1

        # verify that the text is valid as per grammar by compiling it
        # self.compile(text)
        if char_cnt is not None:
            assert len(text) <= char_cnt, f"Text length {len(text)} is greater than the max token count {char_cnt}. Possibly too few characters per token." +\
            f" characters_per_token = {characters_per_token}, max_token_cnt = {max_token_cnt}"
            # text = text[:char_cnt] # Just trim the text from the end because no trimming strategy has worked out
        return text

    def _parse_states_node(self, states_node, context: TrainingDataFormat):
        if len(states_node) >= 1 and states_node[0] == ProofModelGrammar.Keywords.DESCRIPTION.value:
            context.goal_description = states_node[1].strip() if states_node[1] is not None else None
            context.end_goals = states_node[2]
        else:
            context.end_goals = states_node

    def _parse_expr(self, rule_type, nodes, context: TrainingDataFormat):
        if rule_type == ProofModelGrammar.Keywords.HYPOTHESIS:
            hyps = []
            for node in nodes:
                hyps.extend(node)
            return hyps
        elif rule_type == ProofModelGrammar.Keywords.HYPOTHESES:
            return nodes[1] if len(nodes) > 1 else []
        elif rule_type == ProofModelGrammar.Keywords.STATE:
            hyps = nodes[2] if len(nodes) > 2 else []
            goal = nodes[1].strip() if len(nodes) > 1 else None
            state = Goal(hypotheses=hyps, goal=goal)
            return [state]
        elif rule_type == ProofModelGrammar.Keywords.STATES:
            states = []
            for node in nodes:
                states.extend(node)
            return states
        elif rule_type == ProofModelGrammar.Keywords.PROOFSTEP:
            context.proof_steps = [nodes[1]]
            return None
        elif rule_type == "Prog":
            node_idx = 0
            while node_idx < len(nodes):
                if nodes[node_idx] == ProofModelGrammar.Keywords.STATES.value:
                    states_node = nodes[node_idx + 1]
                    self._parse_states_node(states_node, context)
                    break
                node_idx += 1
            if node_idx == len(nodes):
                context.end_goals = nodes
            return context
        else:
            raise Exception("Not implemented")

    def get_action(self, tdf):
        context = tdf
        actions = {
            "Prog": lambda _, nodes: self._parse_expr("Prog", nodes, context),
            "States": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.STATES, nodes, context),
            "State": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.STATE, nodes, context),
            "ProofStep": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.PROOFSTEP, nodes, context),
            "HypsResponses": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.HYPOTHESES, nodes, context),
            "HypResponses": lambda _, nodes: self._parse_expr(ProofModelGrammar.Keywords.HYPOTHESIS, nodes, context),
            "HypResponse": lambda _, nodes: [nodes[1]],
            "String": lambda _, nodes: str(nodes).strip(),
        }
        return actions

    def interpret_result(self, result):
        return result
    
    def parse(self, text, tdf: TrainingDataFormat = None) -> TrainingDataFormat:
        tdf = TrainingDataFormat() if tdf is None else copy.deepcopy(tdf)
        final_tdf : TrainingDataFormat = self.run(text, tdf)
        if self.state_delta:
            final_tdf.end_goals = get_full_state(tdf.start_goals, final_tdf.end_goals)
        return final_tdf

if __name__ == "__main__":
    next_proof_state = """
[STATES]
[DESCRIPTION]
succesfully added the hypothesis to the context
[STATE] 0
algb_mul t (algb_add e c) = algb_add (algb_mul t e) (algb_mul t c)
[HYPOTHESES]
[HYPOTHESIS] c : algb
[HYPOTHESIS] c : algb
[STATE] 1
algb_mul t (algb_add t c) = algb_add (algb_mul t t) (algb_mul t c)
[HYPOTHESES]
[HYPOTHESIS] c : algb
[HYPOTHESIS] t : algb
[END]
"""
    next_proof_state_res = ProofModelPredGrammar().parse(next_proof_state)
    print(next_proof_state_res)

    start_proof_state = """
[STATES]
[DESCRIPTION]
succesfully added the hypothesis to the context
[STATE] 0
algb_add a e = a
[HYPOTHESES]
[HYPOTHESIS] a : algb
[STATE] 1
algb_add e a = a
[HYPOTHESES]
[HYPOTHESIS] a : algb

[PROOFSTEP]
apply algb_add_comm.
[END]
"""
    start_proof_state_res = ProofModelGrammar().parse(start_proof_state)
    print(start_proof_state_res)

    delta_proof_state = """
[STATES]
[DESCRIPTION]
succesfully added the hypothesis to the context
[STATE] 0
DITTO 1
[END]
"""
    next_proof_state_res = ProofModelPredGrammar(state_delta=True).parse(delta_proof_state, start_proof_state_res)
    print(next_proof_state_res)
    print(ProofModelGrammar().format_as_per_grammar(start_proof_state_res))
    print()
    print(ProofModelPredGrammar().format_as_per_grammar(next_proof_state_res))
    actual_proof_state = """
[STATES]
[STATE] 0
is_not_temp s -> is_not_temp d -> is_not_temp d
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 1
In (s, d) (mu1 ++ (s, d) :: mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 2
move_no_temp nil
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 3
is_path ((s, d) :: nil)
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 4
(is_mill mu1 /\ is_mill mu2 /\ dests_disjoint mu1 mu2) /\
((is_mill sigma /\ ~ In d (dests sigma)) /\
 r <> d /\ ~ In r (dests sigma)) /\
((dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\ ~ In r (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In d (dests mu2)) /\ ~ In r (dests mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 5
move_no_temp (mu1 ++ mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 6
temp_last ((d, r) :: (s, d) :: sigma)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 7
is_path ((d, r) :: (s, d) :: sigma)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 8
is_mill mu /\
(is_mill sigma /\
 (is_mill nil /\ ~ In d (dests nil)) /\
 dests_disjoint sigma nil /\ ~ In d (dests sigma)) /\
dests_disjoint mu sigma /\ dests_disjoint mu nil /\ ~ In d (dests mu)
[HYPOTHESES]
[HYPOTHESIS] D : is_path (sigma ++ (s, d) :: nil)
[HYPOTHESIS] C : temp_last (sigma ++ (s, d) :: nil)
[HYPOTHESIS] B : move_no_temp mu
[HYPOTHESIS] A : is_mill mu /\
(is_mill sigma /\
 (is_mill nil /\ ~ In d (dests nil)) /\
 dests_disjoint sigma nil /\ ~ In d (dests sigma)) /\
dests_disjoint mu sigma /\ dests_disjoint mu nil /\ ~ In d (dests mu)
[HYPOTHESIS] WF : state_wf (State mu (sigma ++ (s, d) :: nil) tau)
[HYPOTHESIS] tau : moves

[PROOFSTEP]
auto.
[END]
"""
    state_res = ProofModelGrammar().parse(actual_proof_state)
    next_state = """
[STATES]
[STATE] 0
In (s, d) (mu1 ++ (s, d) :: mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 1
move_no_temp nil
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 2
is_path ((s, d) :: nil)
[HYPOTHESES]
[HYPOTHESIS] D : is_path nil
[HYPOTHESIS] C : temp_last nil
[HYPOTHESIS] B : move_no_temp (mu1 ++ (s, d) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In d (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In d (dests mu1)) /\
is_mill nil /\
dests_disjoint mu1 nil /\ dests_disjoint mu2 nil /\ ~ In d (dests nil)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (s, d) :: mu2) nil tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] s,d : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 3
(is_mill mu1 /\ is_mill mu2 /\ dests_disjoint mu1 mu2) /\
((is_mill sigma /\ ~ In d (dests sigma)) /\
 r <> d /\ ~ In r (dests sigma)) /\
((dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\ ~ In r (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In d (dests mu2)) /\ ~ In r (dests mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 4
move_no_temp (mu1 ++ mu2)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 5
temp_last ((d, r) :: (s, d) :: sigma)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 6
is_path ((d, r) :: (s, d) :: sigma)
[HYPOTHESES]
[HYPOTHESIS] D : is_path ((s, d) :: sigma)
[HYPOTHESIS] C : temp_last ((s, d) :: sigma)
[HYPOTHESIS] B : move_no_temp (mu1 ++ (d, r) :: mu2)
[HYPOTHESIS] A : (is_mill mu1 /\
 (is_mill mu2 /\ ~ In r (dests mu2)) /\
 dests_disjoint mu1 mu2 /\ ~ In r (dests mu1)) /\
(is_mill sigma /\ ~ In d (dests sigma)) /\
(dests_disjoint mu1 sigma /\ ~ In d (dests mu1)) /\
(dests_disjoint mu2 sigma /\ ~ In r (dests sigma)) /\
d <> r /\ ~ In d (dests mu2)
[HYPOTHESIS] WF : state_wf (State (mu1 ++ (d, r) :: mu2) ((s, d) :: sigma) tau)
[HYPOTHESIS] tau : moves
[HYPOTHESIS] sigma : list (reg * reg)
[HYPOTHESIS] s : reg
[HYPOTHESIS] mu2 : list (reg * reg)
[HYPOTHESIS] d,r : reg
[HYPOTHESIS] mu1 : list (reg * reg)
[HYPOTHESIS] val : Type
[HYPOTHESIS] temp : reg -> reg
[HYPOTHESIS] reg_eq : forall r1 r2 : reg, {r1 = r2} + {r1 <> r2}
[HYPOTHESIS] reg : Type
[STATE] 7
is_mill mu /\
(is_mill sigma /\
 (is_mill nil /\ ~ In d (dests nil)) /\
 dests_disjoint sigma nil /\ ~ In d (dests sigma)) /\
dests_disjoint mu sigma /\ dests_disjoint mu nil /\ ~ In d (dests mu)
[HYPOTHESES]
[HYPOTHESIS] D : is_path (sigma ++ (s, d) :: nil)
[HYPOTHESIS] C : temp_last (sigma ++ (s, d) :: nil)
[HYPOTHESIS] B : move_no_temp mu
[HYPOTHESIS] A : is_mill mu /\
(is_mill sigma /\
 (is_mill nil /\ ~ In d (dests nil)) /\
 dests_disjoint sigma nil /\ ~ In d (dests sigma)) /\
dests_disjoint mu sigma /\ dests_disjoint mu nil /\ ~ In d (dests mu)
[HYPOTHESIS] WF : state_wf (State mu (sigma ++ (s, d) :: nil) tau)
[HYPOTHESIS] tau : moves
[END]
"""
    next_state_res = ProofModelPredGrammar().parse(next_state, state_res)
    formatted_out = ProofModelPredGrammar(state_delta=True).format_as_per_grammar(next_state_res)

    actual_proof_state = """
[STATES]
[STATE] 0
forall n m : nat, S n + m = S (n + m)
[END]
"""
    state_res = ProofModelGrammar().parse(actual_proof_state)
    print(state_res)
    next_state = """
[PROOFSTEP]
intros.

[STATES]
[STATE] 0
S n + m = S (n + m)
[HYPOTHESES]
[HYPOTHESIS] n,m : nat
[END]
"""
    next_state_res = ProofModelPredGrammar(state_delta=True).parse(next_state, state_res)
    formatted_out = ProofModelPredGrammar(state_delta=True).format_as_per_grammar(next_state_res)
    print(formatted_out)
    print(next_state_res)


import subprocess
import os
import datetime
import logging
import requests
import time

class TGIServer:
    def __init__(
        self,
        volume: str = None,
        port: int = 8080,
        gpus: bool = True,
        shm_size: str = "1g",
        trust_remote_code: bool = False,
        model_id: str = None,
        path_to_local_model: str = None,
        docker_image: str = "ghcr.io/huggingface/text-generation-inference:latest",
        logdir: str = ".logs/tgi_server",
        hostname: str = "0.0.0.0",
        cuda_visible_devices: str = "0"
    ):
        """
        Initialize TGI Server Parameters.
        
        Args:
            model_id (str): Hugging Face Hub model ID.
            path_to_local_model (str): Local path to the model directory.
            volume (str): Directory to mount into the container, optional if local model is provided.
            port (int): Port to expose the server on (default 8080).
            gpus (bool): Use GPU support if available (default True).
            shm_size (str): Shared memory size (default '1g').
            trust_remote_code (bool): Allow remote custom model code (default False).
            docker_image (str): Docker image for TGI (default HF's TGI image).
            logdir (str): Directory where logs will be stored.
        """
        assert model_id or path_to_local_model, "Either model_id or path_to_local_model must be provided"
        assert not (model_id and path_to_local_model), "Only one of model_id or path_to_local_model can be provided"
        if path_to_local_model:
            assert volume is None or os.path.abspath(volume) == os.path.dirname(path_to_local_model), (
                "If local model path is provided, volume must match its base directory or be omitted."
            )
        
        self.model_id = model_id if model_id else ""
        self.volume = os.path.abspath(volume) if volume else None
        self.port = port
        self.gpus = gpus
        self.shm_size = shm_size
        self.trust_remote_code = trust_remote_code
        self.path_to_local_model = path_to_local_model if path_to_local_model else ""
        self.is_local_model = path_to_local_model is not None
        self.docker_image = docker_image
        self.hostname = hostname
        self.logdir = os.path.abspath(logdir)
        os.makedirs(self.logdir, exist_ok=True)
        self.timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.logdir = os.path.join(self.logdir, self.timestamp)
        os.makedirs(self.logdir, exist_ok=True)
        self.container_id = None
        self.process = None
        self.cuda_visible_devices = cuda_visible_devices

        # Initialize logger
        log_filename = os.path.join(self.logdir, f"tgi_server.log")
        self.logger = logging.getLogger("TGIServer")
        self.logger.setLevel(logging.INFO)
        file_handler = logging.FileHandler(log_filename)
        file_handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
        self.logger.addHandler(file_handler)
        self.logger.info("TGI Server initialized")

    def __enter__(self):
        """Context management: Start the server when entering the context."""
        self.start_server()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context management: Stop the server when exiting the context."""
        self.stop_server()
        if exc_type:
            self.logger.error(f"Exception occurred: {exc_val}")

    def _find_existing_container(self):
        """Find an existing container running on the specified port."""
        try:
            check_command = f"docker ps --filter 'publish={self.port}' --format '{{{{.ID}}}}'"
            result = subprocess.run(check_command, shell=True, check=True, stdout=subprocess.PIPE, text=True)
            container_id = result.stdout.strip()
            return container_id if container_id else None
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Error checking for existing container: {e}")
            return None

    def start_server(self):
        """Build the Docker command and start the TGI server, logging output to a file."""
        # Check if a container is already running on the specified port
        existing_container = self._find_existing_container()
        if existing_container:
            self.logger.info(f"Found existing container {existing_container} running on port {self.port}. Attaching to it.")
            self.container_id = existing_container
            return

        # Validate inputs
        if self.is_local_model and not self.path_to_local_model:
            self.logger.error("Path to the local model is required if is_local_model=True")
            raise ValueError("Path to the local model is required if is_local_model=True")
        
        # Handle local model path with correct volume mapping
        if self.is_local_model:
            base_volume = os.path.dirname(self.path_to_local_model)
            relative_model_path = os.path.relpath(self.path_to_local_model, base_volume)
            model_option = f"--model-id /data/{relative_model_path}"
            volume_mapping = f"-v {base_volume}:/data"
        else:
            model_option = f"--model-id {self.model_id}"
            volume_mapping = f"-v {self.volume}:/data" if self.volume else ""

        # Add trust-remote-code flag if required
        trust_remote_flag = "--trust-remote-code" if self.trust_remote_code else ""

        # Add GPU flag
        gpu_flag = "--gpus all" if self.gpus else ""

        # Set CUDA_VISIBLE_DEVICES environment variable
        environ = f"--env CUDA_VISIBLE_DEVICES={self.cuda_visible_devices}"

        # Build Docker command
        docker_command = (
            f"docker run {environ} {gpu_flag} --shm-size {self.shm_size} -p {self.port}:80 --hostname {self.hostname} "
            f"{volume_mapping} {self.docker_image} "
            f"{model_option} {trust_remote_flag}"
        )
        
        self.logger.info(f"Starting TGI server with command at {datetime.datetime.now().isoformat()}:")
        self.logger.info(docker_command)

        # Run the Docker command and redirect output to log file
        try:
            log_file_path = os.path.join(self.logdir, "docker_output.log")
            with open(log_file_path, "w") as log_file:
                self.process = subprocess.Popen(
                    docker_command,
                    shell=True,
                    stdout=log_file,
                    stderr=subprocess.STDOUT,
                    text=True
                )
                self.logger.info(f"Docker process started with PID: {self.process.pid}")

            # Wait for server readiness by monitoring logs
            if self._wait_for_log_message("Connected", log_file_path):
                # Fetch the container ID
                inspect_command = "docker ps -q -l"
                result = subprocess.run(inspect_command, shell=True, check=True, stdout=subprocess.PIPE, text=True)
                self.container_id = result.stdout.strip()
                self.logger.info(f"TGI server started successfully. Container ID: {self.container_id}")
            else:
                self.logger.error("TGI server failed to start within the timeout period.")
                self.stop_server()
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Error starting TGI server. Check logs at {log_file_path}")
            self.stop_server()
        except Exception as e:
            self.logger.error(f"Unexpected error: {e}")
            self.stop_server()

    def _wait_for_log_message(self, message, log_file_path, timeout=60):
        """Wait for a specific log message to appear in the Docker output log."""
        self.logger.info(f"Waiting for log message: '{message}'")
        start_time = datetime.datetime.now()
        while (datetime.datetime.now() - start_time).seconds < timeout:
            try:
                with open(log_file_path, "r") as log_file:
                    logs = log_file.read()
                    if message in logs:
                        self.logger.info(f"Log message '{message}' found.")
                        return True
            except FileNotFoundError:
                self.logger.warning("Log file not found yet, retrying...")
            time.sleep(1)
        self.logger.error(f"Timeout reached while waiting for log message: '{message}'")
        return False

    def stop_server(self):
        """Stop the running TGI server container."""
        if not self.container_id:
            self.logger.warning("No container ID found. Cannot stop server.")
            return

        stop_command = f"docker stop {self.container_id}"
        self.logger.info(f"Stopping TGI server with command: {stop_command}")

        try:
            subprocess.run(stop_command, shell=True, check=True)
            self.logger.info("TGI server stopped successfully.")
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Error stopping TGI server: {e}")
        except Exception as e:
            self.logger.error(f"Unexpected error: {e}")

        if self.process:
            self.process.terminate()
            self.logger.info(f"Terminated Docker process with PID: {self.process.pid}")

    def is_alive(self):
        """Check if the TGI server is running and responding."""
        if self.container_id:
            self.logger.info(f"Checking if TGI server in container {self.container_id} is alive.")
            # Use ping_test to verify responsiveness
            response = self.ping_test()
            if response is not None:
                self.logger.info("TGI server is alive and responding.")
                return True
            else:
                self.logger.warning("TGI server is running but not responding.")
                return False
        else:
            self.logger.warning("No container ID found. TGI server is not running.")
            return False

    def ping_test(self, retries=3, delay=2):
        """Run a test generation request to verify the TGI server is responsive."""
        url = f"http://{self.hostname}:{self.port}/v1/completions"
        payload = {
            "max_tokens": 32,
            "n": 2,
            "temperature": 0.7,
            "top_p": 0.9,
            "frequency_penalty": 0.0,
            "prompt": "Explain neural networks."
        }
        for attempt in range(1, retries + 1):
            try:
                self.logger.info(f"Sending ping test to {url}, attempt {attempt}/{retries}")
                response = requests.post(url, json=payload, headers={"Content-Type": "application/json"})
                if response.status_code == 200:
                    self.logger.info("Ping test successful. Response: %s", response.json())
                    return response.json()
                else:
                    self.logger.error(f"Ping test failed with status code {response.status_code}. Response: {response.text}")
            except requests.ConnectionError as e:
                self.logger.error(f"Ping test failed. Connection error: {e}")
            except Exception as e:
                self.logger.error(f"Unexpected error during ping test: {e}")
            self.logger.info(f"Retrying ping test after {delay} seconds...")
            time.sleep(delay)
        self.logger.error("Ping test failed after all retries.")
        return None

if __name__ == "__main__":
    model_path = '<root>/models/proof-wala-codet5-small-coq-lean-4-2048/checkpoint-337500'
    if "<root>" in model_path:
        model_path = model_path.replace("<root>/", os.environ.get("ROOT", "").trim('/') + "/")
    with TGIServer(
        path_to_local_model=model_path,
        port=8080,
        gpus=True,
        logdir=".logs/tgi_server"
    ) as tgi_server:
        if tgi_server.is_alive():
            tgi_server.ping_test()


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
from proof_wala.llm_helpers.model import Model
from proof_wala.llm_helpers.llama2_chat_format import Llama2ChatFormat
from proof_wala.parsers.grammars.prompt_template_grammar import PromptGrammar


class LlamaChatBot(object):
    def __init__(self, 
        bot_name: str, 
        model_name: str, 
        system_prompt_filepath: str = None, 
        sample_conversation_filepath: str = None, 
        model_args: dict = None):
        self.bot_name = bot_name
        self.model_name = model_name
        self.system_prompt_filepath = system_prompt_filepath
        self.sample_conversation_filepath = sample_conversation_filepath
        self.model_args = model_args
        self._system_prompt = None
        self._sample_conversation = None
        self._prompt_grammar = PromptGrammar()
        self._messages = []
        self._chat_formatter = Llama2ChatFormat()
        pass

    def load(self):
        self.model = Model(self.model_name, training_args=None, log_folder=None, **self.model_args)
        self.model.load()
        self._system_prompt = self._prompt_grammar.get_main_message(self.system_prompt_filepath)
        self._sample_conversation = self._prompt_grammar.get_conv_messages(self.sample_conversation_filepath)
        self._messages.append(self._system_prompt)
        self._messages += self._sample_conversation
    
    def __call__(self, message: str) -> str:
        message = {"role": "user", "content": message}
        self._messages.append(message)
        self._respond()
        return self._messages[-1]["content"]
    
    def _respond(self) -> str:
        # Format the messages
        message, _ = self._chat_formatter(self._messages)
        # Call the model
        generated_response = self.model.generate(
                message,
                max_new_tokens=200,
                temperature=0.1, # Nucleus sampling
                do_sample=True, # Nucleus sampling
                top_k=5, # Nucleus sampling
                # num_beams=5, # Beam search
                num_return_sequences=1,
                stop_tokens=[self.model._tokenizer.eos_token],
                padding=True,
                #truncation=True,
                return_full_text=False,
                skip_special_tokens=True)
        response = generated_response[0].generated_text[0]
        self._messages.append({"role": "assistant", "content": response})

class LlamaConsoleChatBot(object):
    def __init__(self,
        bot_name: str, 
        model_name: str, 
        system_prompt_filepath: str = None, 
        sample_conversation_filepath: str = None, 
        model_args: dict = None):
        self.llama_chat_bot = LlamaChatBot(bot_name, model_name, system_prompt_filepath, sample_conversation_filepath, model_args)
        pass

    def load(self):
        self.llama_chat_bot.load()
        pass

    def run(self):
        while True:
            message = input(f"You>> ")
            if message == "exit":
                break
            response = self.llama_chat_bot(message)
            response = f"{self.llama_chat_bot.bot_name}>> {response}"
            print(response)
        pass

if __name__ == "__main__":
    import json
    model_name = "meta-llama/Llama-2-7b-chat-hf"
    with open(".secrets/huggingface_token.json", "r") as f:
        hf_token = json.load(f)["token"]
    chatbot_name = "llama_chat_bot"
    problem_name = "simple_math"
    system_prompt_filepath = f"src/proof_wala/data/prompts/system/{problem_name}.md"
    sample_conversation_filepath = f"src/proof_wala/data/prompts/conversation/{problem_name}.md"
    console_chatbot = LlamaConsoleChatBot(
        chatbot_name, 
        model_name, 
        system_prompt_filepath, 
        sample_conversation_filepath, 
        model_args={"token": hf_token})
    console_chatbot.load()
    console_chatbot.run()